import codecs
import os
import platform
import re
import shutil
import stat
import subprocess
import sys
import datetime

from setuptools import setup, Command
from pkg_resources import get_build_platform
from wheel.bdist_wheel import bdist_wheel as bdist_wheel_


NAME = "spirit"
PACKAGES = ["spirit", "spirit.parameters"]
META_PATH = os.path.join("spirit", "__init__.py")
KEYWORDS = ["Spirit", "Spin Dynamics"]
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Science/Research",
    "Natural Language :: English",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Programming Language :: C",
    "Programming Language :: C++",
    "Programming Language :: Python",
    "Topic :: Scientific/Engineering",
    "Topic :: Software Development :: Libraries :: Python Modules",
]
INSTALL_REQUIRES = ["numpy"]

###############################################################################

HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file. Assume UTF-8 encoding, but replace Windows CRLF with Unix LF.
    """
    with codecs.open(os.path.join(HERE, *parts), "rb", "utf-8") as f:
        return f.read().replace("\r\n", "\n")


META_FILE = read(META_PATH)


def find_meta(meta):
    """
    Extract __*meta*__ from META_FILE.
    """
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta), META_FILE, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError("Unable to find __{meta}__ string.".format(meta=meta))


def get_git_commit_datetime():
    try:
        commit_hash = (
            subprocess.check_output(
                "git rev-parse HEAD", shell=True, stderr=subprocess.STDOUT
            )
            .decode("utf-8")
            .strip()
        )
        commit_datetime = (
            subprocess.check_output(
                "git show -s --format=%ci " + commit_hash,
                shell=True,
                stderr=subprocess.STDOUT,
            )
            .decode("utf-8")
            .strip()
        )
        print(commit_datetime)
        commit_datetime = " ".join(commit_datetime.split()[:-1])
        print(commit_datetime)
        datetime_object = datetime.datetime.strptime(
            commit_datetime, "%Y-%m-%d %H:%M:%S"
        )
        print("{:%Y%m%d%H%M%S}".format(datetime_object))
        return "{:%Y%m%d%H%M%S}".format(datetime_object)
    except subprocess.CalledProcessError as cpe:
        print(cpe.output)
        return "00000000000000"


import unittest


def my_test_suite():
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover("test", pattern="*.py")
    return test_suite


class bdist_wheel(bdist_wheel_):
    def finalize_options(self):
        from sys import platform as _platform

        platform_name = get_build_platform()
        if _platform == "linux" or _platform == "linux2":
            # Linux
            platform_name = "manylinux1_x86_64"

        bdist_wheel_.finalize_options(self)
        self.universal = True
        self.plat_name_supplied = True
        self.plat_name = platform_name


class CleanCommand(Command):
    """Custom clean command to tidy up the project root."""

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.system("rm -vrf ./build ./dist ./*.pyc ./*.tgz ./*.egg-info")


if __name__ == "__main__":
    # If the environment variable SPIRIT_VERSION_SUFFIX is defined,
    # it is appended to the package version number.
    version_suffix = ""
    add_version_suffix = os.environ.get("SPIRIT_ADD_VERSION_SUFFIX", "")
    if add_version_suffix.lower() in ("yes", "true", "t", "1"):
        timepoint_string = get_git_commit_datetime()
        if timepoint_string == "00000000000000":
            timepoint_string = "{:%Y%m%d%H%M}".format(datetime.datetime.now())
        version_suffix = ".dev" + timepoint_string
        print("setup.py: package version suffix = ", version_suffix)

    # Setup the package info
    setup(
        name=NAME,
        description=find_meta("description"),
        long_description=read("README.md"),
        long_description_content_type="text/markdown",
        license=find_meta("license"),
        url=find_meta("uri"),
        version=find_meta("version") + version_suffix,
        author=find_meta("author"),
        author_email=find_meta("email"),
        maintainer=find_meta("author"),
        maintainer_email=find_meta("email"),
        keywords=KEYWORDS,
        packages=PACKAGES,
        classifiers=CLASSIFIERS,
        install_requires=INSTALL_REQUIRES,
        package_data={
            "spirit": ["libSpirit.dylib", "libSpirit.so", "Spirit.dll"],
        },
        cmdclass={"bdist_wheel": bdist_wheel, "clean": CleanCommand},
        test_suite="setup.my_test_suite",
    )



"""
System
====================
"""

from spirit import spiritlib
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()

from spirit.scalar import scalar
from spirit import parameters
from numpy import frombuffer, ndarray as np

### Get Chain index
_Get_Index = _spirit.System_Get_Index
_Get_Index.argtypes = [ctypes.c_void_p]
_Get_Index.restype = ctypes.c_int


def get_index(p_state):
    """Returns the index of the currently active image."""
    return int(_Get_Index(ctypes.c_void_p(p_state)))


### Get Chain number of images
_Get_NOS = _spirit.System_Get_NOS
_Get_NOS.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_NOS.restype = ctypes.c_int


def get_nos(p_state, idx_image=-1, idx_chain=-1):
    """Returns the number of spins (NOS)."""
    return int(
        _Get_NOS(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )


### Get Pointer to Spin Directions
# NOTE: Changing the values of the array_view one can alter the value of the data of the state
_Get_Spin_Directions = _spirit.System_Get_Spin_Directions
_Get_Spin_Directions.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_Spin_Directions.restype = ctypes.POINTER(scalar)


def get_spin_directions(p_state, idx_image=-1, idx_chain=-1):
    """Returns an `numpy.array_view` of shape (NOS, 3) with the components of each spins orientation vector.

    Changing the contents of this array_view will have direct effect on calculations etc.
    """
    nos = get_nos(p_state, idx_image, idx_chain)
    ArrayType = scalar * 3 * nos
    Data = _Get_Spin_Directions(
        ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
    )
    array_pointer = ctypes.cast(Data, ctypes.POINTER(ArrayType))
    array = frombuffer(array_pointer.contents, dtype=scalar)
    array_view = array.view()
    array_view.shape = (nos, 3)
    return array_view


### Get Pointer to Effective Field
# NOTE: Changing the values of the array_view one can alter the value of the data of the state
_Get_Effective_Field = _spirit.System_Get_Effective_Field
_Get_Effective_Field.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_Effective_Field.restype = ctypes.POINTER(scalar)


def get_effective_field(p_state, idx_image=-1, idx_chain=-1):
    nos = get_nos(p_state, idx_image, idx_chain)
    ArrayType = scalar * 3 * nos
    Data = _Get_Effective_Field(
        ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
    )
    array_pointer = ctypes.cast(Data, ctypes.POINTER(ArrayType))
    array = frombuffer(array_pointer.contents, dtype=scalar)
    array_view = array.view()
    array_view.shape = (nos, 3)
    return array_view


### Get Pointer to an eigenmode
# NOTE: Changing the values of the array_view one can alter the value of the data of the state
_Get_Eigenmode = _spirit.System_Get_Eigenmode
_Get_Eigenmode.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int, ctypes.c_int]
_Get_Eigenmode.restype = ctypes.POINTER(scalar)


def get_eigenmode(p_state, idx_mode, idx_image=-1, idx_chain=-1):
    nos = get_nos(p_state, idx_image, idx_chain)
    ArrayType = scalar * 3 * nos
    Data = _Get_Eigenmode(
        ctypes.c_void_p(p_state),
        ctypes.c_int(idx_mode),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )
    array_pointer = ctypes.cast(Data, ctypes.POINTER(ArrayType))
    array = frombuffer(array_pointer.contents, dtype=scalar)
    array_view = array.view()
    array_view.shape = (nos, 3)
    return array_view


### Get total Energy
_Get_Energy = _spirit.System_Get_Energy
_Get_Energy.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Get_Energy.restype = ctypes.c_float


def get_energy(p_state, idx_image=-1, idx_chain=-1):
    """Calculates and returns the energy of the system."""
    return float(
        _Get_Energy(
            ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
        )
    )


### Get Energy
_Get_Eigenvalues = _spirit.System_Get_Eigenvalues
_Get_Eigenvalues.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
    ctypes.c_int,
]
_Get_Eigenvalues.restype = None


def get_eigenvalues(p_state, idx_image=-1, idx_chain=-1):
    noe = parameters.ema.getNModes(p_state, idx_image, idx_chain)
    eigenvalues = (noe * ctypes.c_float)()
    _Get_Eigenvalues(
        ctypes.c_void_p(p_state),
        eigenvalues,
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )
    return eigenvalues


### Get Energy Contributions
### The result is a dictionary with strings as keys and floats as values
### The keys are the names of the energy contributions, the values the energy_contribution in meV
_Get_Energy_Array = _spirit.System_Get_Energy_Array
_Get_Energy_Array.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_bool,
    ctypes.c_int,
    ctypes.c_int,
]
_Get_Energy_Array.restype = None

_Get_Energy_Array_Names = _spirit.System_Get_Energy_Array_Names
_Get_Energy_Array_Names.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_char),
    ctypes.c_int,
    ctypes.c_int,
]
_Get_Energy_Array_Names.restype = ctypes.c_int


def get_energy_contributions(
    p_state, divide_by_nspins=True, idx_image=-1, idx_chain=-1
):
    NULL = ctypes.POINTER(ctypes.c_char)()

    n_char_array = _Get_Energy_Array_Names(
        ctypes.c_void_p(p_state), NULL, ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
    )

    energy_array_names = (n_char_array * ctypes.c_char)()

    _Get_Energy_Array_Names(
        ctypes.c_void_p(p_state),
        energy_array_names,
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )

    contrib_names = str(energy_array_names[:].decode("utf-8")).split("|")
    n_contribs = len(contrib_names)
    energies = (n_contribs * ctypes.c_float)()

    _Get_Energy_Array(
        ctypes.c_void_p(p_state),
        energies,
        divide_by_nspins,
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )

    return dict(zip(contrib_names, energies))


### Get Chain number of images
_Update_Data = _spirit.System_Update_Data
_Update_Data.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Update_Data.restype = None


def update_data(p_state, idx_image=-1, idx_chain=-1):
    """TODO: document when this needs to be called."""
    _Update_Data(
        ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
    )


### Eigenmodes
_Eigenmodes = _spirit.System_Update_Eigenmodes
_Eigenmodes.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Eigenmodes.restype = None


def update_eigenmodes(p_state, idx_image=-1, idx_chain=-1):
    """Calculates eigenmodes of a system according to EMA parameters.
    This needs to be called or eigenmodes need to be read in before they can be used by other functions
    (e.g. writing them to a file).
    """
    spiritlib.wrap_function(
        _Eigenmodes,
        [ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)],
    )


### Print Energy array
_Print_Energy_Array = _spirit.System_Print_Energy_Array
_Print_Energy_Array.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
_Print_Energy_Array.restype = None


def print_energy_array(p_state, idx_image=-1, idx_chain=-1):
    """Print the energy array of the state to the console."""
    _Print_Energy_Array(
        ctypes.c_void_p(p_state), ctypes.c_int(idx_image), ctypes.c_int(idx_chain)
    )



"""
Log
====================
"""

from spirit import spiritlib
import ctypes

### Load Library
_spirit = spiritlib.load_spirit_library()

# Log levels
LEVEL_ALL = 0
"""Only log highest-level messages, as well as `Severe` and `Error` messages."""

LEVEL_SEVERE = 1
"""Only log highest-level messages, as well as `Severe` and `Error` messages."""

LEVEL_ERROR = 2
"""Only log highest-level messages, as well as `Severe` and `Error` messages."""

LEVEL_WARNING = 3
"""Accept log messages of `Warning` and higher levels."""

LEVEL_PARAMETER = 4
"""Accept log messages of `Parameter` and higher levels."""

LEVEL_INFO = 5
"""Accept log messages of `Info` and higher levels."""

LEVEL_DEBUG = 6
"""Accept log messages of `Debug` and higher levels."""

# Log message senders
SENDER_ALL = 0
"""General message sender."""

SENDER_IO = 1
"""Everything related to I/O."""

SENDER_GNEB = 2
"""GNEB method message sender."""

SENDER_LLG = 3
"""LLG method message sender."""

SENDER_MC = 4
"""Monte Carlo method message sender."""

SENDER_MMF = 5
"""MMF method message sender."""

SENDER_API = 6
"""API message sender."""

SENDER_UI = 7
"""User interface message sender."""

_Send = _spirit.Log_Send
_Send.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_char_p,
    ctypes.c_int,
    ctypes.c_int,
]
_Send.restype = None


def send(p_state, level, sender, message, idx_image=-1, idx_chain=-1):
    """Add a message to the log.

    Arguments:

    - `level`: see integers defined above. The message may be printed to the console and/or written to the log file, depending on the current log parameters
    - `sender`: see integers defined above. Used to distinguish context
    - `message`: a string which to log

    Keyword arguments:

    - `idx_image`: can be used to specify to which image the message relates (default: active image)
    """
    _Send(
        ctypes.c_void_p(p_state),
        ctypes.c_int(level),
        ctypes.c_int(sender),
        ctypes.c_char_p(message.encode("utf-8")),
        ctypes.c_int(idx_image),
        ctypes.c_int(idx_chain),
    )


_Append = _spirit.Log_Append
_Append.argtypes = [ctypes.c_void_p]
_Append.restype = None


def append(p_state):
    """Force the appending of new messages to the log file."""
    _Append(ctypes.c_void_p(p_state))


_Get_N_Entries = _spirit.Log_Get_N_Entries
_Get_N_Entries.argtypes = [ctypes.c_void_p]
_Get_N_Entries.restype = ctypes.c_int


def get_n_entries(p_state):
    """Returns the number of Log entries."""
    return int(_Get_N_Entries(ctypes.c_void_p(p_state)))


_Get_N_Errors = _spirit.Log_Get_N_Errors
_Get_N_Errors.argtypes = [ctypes.c_void_p]
_Get_N_Errors.restype = ctypes.c_int


def get_n_errors(p_state):
    """Returns the number of error messages that have been logged."""
    return int(_Get_N_Errors(ctypes.c_void_p(p_state)))


_Get_N_Warnings = _spirit.Log_Get_N_Warnings
_Get_N_Warnings.argtypes = [ctypes.c_void_p]
_Get_N_Warnings.restype = ctypes.c_int


def get_n_warnings(p_state):
    """Returns the number of warning messages that have been logged."""
    return int(_Get_N_Warnings(ctypes.c_void_p(p_state)))


_Set_Output_File_Tag = _spirit.Log_Set_Output_File_Tag
_Set_Output_File_Tag.argtypes = [ctypes.c_void_p, ctypes.c_char_p]
_Set_Output_File_Tag.restype = None


def set_output_file_tag(p_state, tag):
    """Set the tagging string which is placed in front of the log file.

    If `"<time>"` is used for the tag, it will be the time of creation of the state.
    """
    _Set_Output_File_Tag(ctypes.c_void_p(p_state), ctypes.c_char_p(tag.encode("utf-8")))


_Set_Output_Folder = _spirit.Log_Set_Output_Folder
_Set_Output_Folder.argtypes = [ctypes.c_void_p, ctypes.c_char_p]
_Set_Output_Folder.restype = None


def set_output_folder(p_state, tag):
    """Set the output folder in which to place the log file."""
    _Set_Output_Folder(ctypes.c_void_p(p_state), ctypes.c_char_p(tag.encode("utf-8")))


_Set_Output_To_Console = _spirit.Log_Set_Output_To_Console
_Set_Output_To_Console.argtypes = [ctypes.c_void_p, ctypes.c_bool, ctypes.c_int]
_Set_Output_To_Console.restype = None


def set_output_to_console(p_state, output, level):
    """Set whether the Log is output to the console and the level up to which messages are logged."""
    _Set_Output_To_Console(
        ctypes.c_void_p(p_state), ctypes.c_bool(output), ctypes.c_int(level)
    )


_Set_Output_To_File = _spirit.Log_Set_Output_To_File
_Set_Output_To_File.argtypes = [ctypes.c_void_p, ctypes.c_bool, ctypes.c_int]
_Set_Output_To_File.restype = None


def set_output_to_file(p_state, output, level):
    """Set whether the Log is output to a file and the level up to which messages are logged."""
    _Set_Output_To_File(
        ctypes.c_void_p(p_state), ctypes.c_bool(output), ctypes.c_int(level)
    )


_Get_Output_To_Console = _spirit.Log_Get_Output_To_Console
_Get_Output_To_Console.argtypes = [ctypes.c_void_p]
_Get_Output_To_Console.restype = ctypes.c_bool


def get_output_to_console(p_state):
    """Returns a bool indicating whether the Log is output to the console."""
    return bool(_Get_Output_To_Console(ctypes.c_void_p(p_state)))


_Get_Output_Console_Level = _spirit.Log_Get_Output_Console_Level
_Get_Output_Console_Level.argtypes = [ctypes.c_void_p]
_Get_Output_Console_Level.restype = ctypes.c_int


def get_output_console_level(p_state):
    """Returns the level up to which the Log is output to the console.

    The return value will be one of the integers defined above.
    """
    return int(_Get_Output_Console_Level(ctypes.c_void_p(p_state)))


_Get_Output_To_File = _spirit.Log_Get_Output_To_File
_Get_Output_To_File.argtypes = [ctypes.c_void_p]
_Get_Output_To_File.restype = ctypes.c_bool


def get_output_to_file(p_state):
    """Returns a bool indicating whether the Log is output to a file."""
    return bool(_Get_Output_To_File(ctypes.c_void_p(p_state)))


_Get_Output_File_Level = _spirit.Log_Get_Output_File_Level
_Get_Output_File_Level.argtypes = [ctypes.c_void_p]
_Get_Output_File_Level.restype = ctypes.c_int


def get_output_file_level(p_state):
    """Returns the level up to which the Log is output to a file.

    The return value will be one of the integers defined above.
    """
    return int(_Get_Output_File_Level(ctypes.c_void_p(p_state)))



"""
HTST
====================

Harmonic transition state theory.

Note that `calculate_prefactor` needs to be called before using any of the getter functions.
"""

from spirit import spiritlib
from spirit import parameters
from spirit import system
import ctypes
import numpy as np

### Load Library
_spirit = spiritlib.load_spirit_library()


_Calculate = _spirit.HTST_Calculate
_Calculate.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_bool,
    ctypes.c_int,
]
_Calculate.restype = ctypes.c_float


def calculate(
    p_state,
    idx_image_minimum,
    idx_image_sp,
    n_eigenmodes_keep=-1,
    sparse=False,
    idx_chain=-1,
):
    """Performs an HTST calculation and returns rate prefactor.

    *Note:* this function must be called before any of the getters.
    """
    return _Calculate(
        p_state, idx_image_minimum, idx_image_sp, n_eigenmodes_keep, sparse, idx_chain
    )


### Get HTST transition rate components
_Get_Info = _spirit.HTST_Get_Info
_Get_Info.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_float),
    ctypes.POINTER(ctypes.c_int),
    ctypes.c_int,
]
_Get_Info.restype = None


def get_info(p_state, idx_chain=-1):
    """Returns a set of HTST information:

    - the exponent of the temperature-dependence
    - me
    - Omega_0
    - s
    - zero mode volume at the minimum
    - zero mode volume at the saddle point
    - dynamical prefactor
    - full rate prefactor (without temperature dependent part)
    """
    temperature_exponent = ctypes.c_float()
    me = ctypes.c_float()
    Omega_0 = ctypes.c_float()
    s = ctypes.c_float()
    volume_min = ctypes.c_float()
    volume_sp = ctypes.c_float()
    prefactor_dynamical = ctypes.c_float()
    prefactor = ctypes.c_float()
    n_eigenmodes_keep = ctypes.c_int()

    _Get_Info(
        ctypes.c_void_p(p_state),
        ctypes.pointer(temperature_exponent),
        ctypes.pointer(me),
        ctypes.pointer(Omega_0),
        ctypes.pointer(s),
        ctypes.pointer(volume_min),
        ctypes.pointer(volume_sp),
        ctypes.pointer(prefactor_dynamical),
        ctypes.pointer(prefactor),
        ctypes.pointer(n_eigenmodes_keep),
        ctypes.c_int(idx_chain),
    )

    return (
        temperature_exponent.value,
        me.value,
        Omega_0.value,
        s.value,
        volume_min.value,
        volume_sp.value,
        prefactor_dynamical.value,
        prefactor.value,
    )


def get_info_dict(p_state, idx_chain=-1):
    """Returns a set of HTST information in a dictionary:

    - the exponent of the temperature-dependence
    - me
    - Omega_0
    - s
    - zero mode volume at the minimum
    - zero mode volume at the saddle point
    - dynamical prefactor
    - full rate prefactor (without temperature dependent part)
    """
    temperature_exponent = ctypes.c_float()
    me = ctypes.c_float()
    Omega_0 = ctypes.c_float()
    s = ctypes.c_float()
    volume_min = ctypes.c_float()
    volume_sp = ctypes.c_float()
    prefactor_dynamical = ctypes.c_float()
    prefactor = ctypes.c_float()
    n_eigenmodes_keep = ctypes.c_int()

    _Get_Info(
        ctypes.c_void_p(p_state),
        ctypes.pointer(temperature_exponent),
        ctypes.pointer(me),
        ctypes.pointer(Omega_0),
        ctypes.pointer(s),
        ctypes.pointer(volume_min),
        ctypes.pointer(volume_sp),
        ctypes.pointer(prefactor_dynamical),
        ctypes.pointer(prefactor),
        ctypes.pointer(n_eigenmodes_keep),
        ctypes.c_int(idx_chain),
    )

    return {
        "temperature_exponent": temperature_exponent.value,
        "me": me.value,
        "Omega_0": Omega_0.value,
        "s": s.value,
        "volume_min": volume_min.value,
        "volume_sp": volume_sp.value,
        "prefactor_dynamical": prefactor_dynamical.value,
        "prefactor": prefactor.value,
        "n_eigenmodes_keep": n_eigenmodes_keep.value,
    }


_Get_Eigenvalues_Min = _spirit.HTST_Get_Eigenvalues_Min
_Get_Eigenvalues_Min.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
]
_Get_Eigenvalues_Min.restype = None


def get_eigenvalues_min(p_state, idx_chain=-1):
    """Returns the eigenvalues at the minimum with `shape(2*nos)`."""
    nos = system.get_nos(p_state, -1, idx_chain)
    eigenvalues_min = (2 * nos * ctypes.c_float)()
    _Get_Eigenvalues_Min(
        ctypes.c_void_p(p_state), eigenvalues_min, ctypes.c_int(idx_chain)
    )
    return eigenvalues_min


_Get_Eigenvectors_Min = _spirit.HTST_Get_Eigenvectors_Min
_Get_Eigenvectors_Min.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
]
_Get_Eigenvectors_Min.restype = None


def get_eigenvectors_min(p_state, idx_chain=-1):
    """Returns a numpy array view to the eigenvectors at the minimum with `shape(n_eigenmodes_keep, 2*nos)`."""
    n_modes = get_info_dict(p_state)["n_eigenmodes_keep"]
    nos = system.get_nos(p_state, -1, idx_chain)
    eigenvectors_min = (2 * nos * n_modes * ctypes.c_float)()

    ArrayType = ctypes.c_float * (2 * nos * n_modes)
    ev_list = [] * (2 * nos * n_modes)
    _ev_buffer = ArrayType(*ev_list)

    _Get_Eigenvectors_Min(ctypes.c_void_p(p_state), _ev_buffer, ctypes.c_int(idx_chain))

    ev_array = np.array(_ev_buffer)
    ev_view = ev_array.view()
    ev_view.shape = (n_modes, 2 * nos)

    return ev_view


_Get_Eigenvalues_SP = _spirit.HTST_Get_Eigenvalues_SP
_Get_Eigenvalues_SP.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
]
_Get_Eigenvalues_SP.restype = None


def get_eigenvalues_sp(p_state, idx_chain=-1):
    """Returns the eigenvalues at the saddle point with `shape(2*nos)`."""
    nos = system.get_nos(p_state, -1, idx_chain)
    eigenvalues_sp = (2 * nos * ctypes.c_float)()
    _Get_Eigenvalues_SP(
        ctypes.c_void_p(p_state), eigenvalues_sp, ctypes.c_int(idx_chain)
    )
    return eigenvalues_sp


_Get_Eigenvectors_SP = _spirit.HTST_Get_Eigenvectors_SP
_Get_Eigenvectors_SP.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
]
_Get_Eigenvectors_SP.restype = None


def get_eigenvectors_sp(p_state, idx_chain=-1):
    """Returns a numpy array view to the eigenvectors at the saddle point with `shape(n_eigenmodes_keep, 2*nos)`."""

    n_modes = get_info_dict(p_state)["n_eigenmodes_keep"]
    nos = system.get_nos(p_state, -1, idx_chain)

    ArrayType = ctypes.c_float * (2 * nos * n_modes)
    ev_list = [] * (2 * nos * n_modes)
    _ev_buffer = ArrayType(*ev_list)

    _Get_Eigenvectors_SP(ctypes.c_void_p(p_state), _ev_buffer, ctypes.c_int(idx_chain))

    ev_array = np.array(_ev_buffer)
    ev_view = ev_array.view()
    ev_view.shape = (n_modes, 2 * nos)

    return ev_view


_Get_Velocities = _spirit.HTST_Get_Velocities
_Get_Velocities.argtypes = [
    ctypes.c_void_p,
    ctypes.POINTER(ctypes.c_float),
    ctypes.c_int,
]
_Get_Velocities.restype = None


def get_velocities(p_state, idx_chain=-1):
    """Returns the velocities perpendicular to the dividing surface with `shape(2*nos)`."""
    nos = system.get_nos(p_state, -1, idx_chain)
    velocities = (2 * nos * ctypes.c_float)()
    _Get_Velocities(ctypes.c_void_p(p_state), velocities, ctypes.c_int(idx_chain))
    return velocities
